% 
% function reconstructed_image = ...
%     MS_BCS_SPL_DDWT_Decoder(y, Phi, ...
%     subrates, num_rows, num_cols, num_ddwt_levels, max_iterations)
% 
%   This function performs SPL reconstruction of y using a DDWT
%   sparsity basis. Phi gives the projection matrix. The reconstructed
%   image, of size num_rows x num_cols, is returned as
%   reconstructed_image.
%
%   The DDWT transform used for reconstruction has num_ddwt_levels
%   decomposition levels and is implemented using the cplxdual2D
%   program from the Wavelet Software package from
%   http://taco.poly.edu/WaveletSoftware/. The SPL reconstruction uses
%   bivariate shrinkage to perform the thresholding step of the
%   Landweber iteration. The Wavelet Software package must be in
%   Matlab's current search path for both the DDWT as well as
%   bivariate shrinkage.
%
%   max_iterations is an optional parameter specifying the number of
%   iterations of SPL reconstruction to run; default = 200.
% 

%
% MS-BCS-SPL: Multiscale Block Compressed Sensing - Smooth Projected Landweber
% Copyright (C) 2012  James E. Fowler
% http://www.ece.mstate.edu/~fowler
% 
% This program is free software; you can redistribute it and/or
% modify it under the terms of the GNU General Public License
% as published by the Free Software Foundation; either version 2
% of the License, or (at your option) any later version.
% 
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU General Public License for more details.
% 
% You should have received a copy of the GNU General Public License
% along with this program; if not, write to the Free Software
% Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
%

function reconstructed_image = ...
    MS_BCS_SPL_DDWT_Decoder(y, Phi, ...
    subrates, num_rows, num_cols, num_ddwt_levels, max_iterations)

num_levels = length(y) - 1;

if (isnumeric(Phi{1}))
  for level = 1:num_levels
    block_sizes(level) = sqrt(size(Phi{level}, 2));
    Phi_matrix = Phi{level};
    Phi_transpose{level} = @(y) Phi_matrix' * y;
    Phi{level} = @(x) Phi_matrix * x;
  end
elseif (iscell(Phi{1}))
  for level = 1:num_levels
    block_sizes(level) = Phi{level}{1};
    Phi_transpose{level} = Phi{level}{3};
    Phi{level} = Phi{level}{2};
  end
else
  disp('Incorrect Phi format');
  return;
end

lambda = 25;
if (nargin == 6)
  max_iterations = 200;
end

TOL = 0.0001;
D_prev = 0;

x_tilde{1} = y{1};
for level = 1:num_levels
  for subband = 1:3
    x_tilde{level + 1}{subband} = Phi_transpose{level}(y{level + 1}{subband});
  end
end

for i = 1:max_iterations
  [x_tilde D] = SPLIteration(y, x_tilde, Phi, Phi_transpose, ...
      subrates, block_sizes, ...
      num_rows, num_cols, lambda, num_levels, num_ddwt_levels);

  if ((D_prev ~= 0) && (abs(D - D_prev) < TOL))
    break;
  end

  D_prev = D;
end

if (max_iterations > 1)
  [x_tilde D] = SPLIteration(y, x_tilde, Phi, Phi_transpose, ...
      subrates, block_sizes, ...
      num_rows, num_cols, lambda, num_levels, num_ddwt_levels);
end

reconstructed_image = InverseBlockDWT(x_tilde);


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


function [x_tilde D] = SPLIteration(y, x_tilde, Phi, Phi_transpose, ...
    subrates, block_sizes, ...
    num_rows, num_cols, lambda, num_levels, num_ddwt_levels)

if (exist('nor_dualtree.mat'))
  load nor_dualtree
else
  normaliz_coefcalc_dual_tree
end

[Faf, Fsf] = AntonB;
[af, sf] = dualfilt1;

x = InverseBlockDWT(x_tilde);
x_hat = wiener2(x, [3, 3]);
x_tilde_hat = BlockDWT(x_hat, block_sizes);

x_tilde_hat{1} = x_tilde{1};
for level = 1:num_levels
  for subband = 1:3
    if (subrates(level) == 1)
      x_tilde_hat{level + 1}{subband} = x_tilde{level + 1}{subband};
    else
      x_tilde_hat{level + 1}{subband} = ...
	  x_tilde_hat{level + 1}{subband} + ...
	  Phi_transpose{level}((y{level + 1}{subband} - ...
	  Phi{level}(x_tilde_hat{level + 1}{subband})));
    end
  end
end

x1 = InverseBlockDWT(x_tilde_hat);

x_check = normcoef(cplxdual2D(symextend(x1, 2^(num_ddwt_levels - 1)), ...
    num_ddwt_levels, Faf, af), num_ddwt_levels, nor);

if (nargin == 9)
  end_level = 1;
else
  end_level = num_ddwt_levels - 1;
end
x_check = SPLBivariateShrinkage(x_check, end_level, lambda);

x_bar = icplxdual2D(unnormcoef(x_check, num_ddwt_levels, nor), ...
    num_ddwt_levels, Fsf, sf);
Irow = (2^(num_ddwt_levels - 1) + 1):(2^(num_ddwt_levels - 1) + num_rows);
Icol = (2^(num_ddwt_levels - 1) + 1):(2^(num_ddwt_levels - 1) + num_cols);
x_bar = x_bar(Irow, Icol);

x_tilde_bar = BlockDWT(x_bar, block_sizes);

x_tilde_bar{1} = x_tilde{1};
for level = 1:num_levels
  for subband = 1:3
    if (subrates(level) == 1)
      x_tilde_bar{level + 1}{subband} = x_tilde{level + 1}{subband};
    else
      x_tilde_bar{level + 1}{subband} = ...
	  x_tilde_bar{level + 1}{subband} + ...
	  Phi_transpose{level}((y{level + 1}{subband} - ...
	  Phi{level}(x_tilde_bar{level + 1}{subband})));
    end
  end
end  

x_tilde = x_tilde_bar;

x2 = InverseBlockDWT(x_tilde);
  
D = RMS(x1, x2);
  

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


function x_check = SPLBivariateShrinkage(x_check, end_level, lambda)

windowsize  = 3;
windowfilt = ones(1, windowsize)/windowsize;

tmp = x_check{1}{1}{1}{1};
Nsig = median(abs(tmp(:)))/0.6745;

for scale = 1:end_level
  for dir = 1:2
    for dir1 = 1:3
      Y_coef_real = x_check{scale}{1}{dir}{dir1};
      Y_coef_imag = x_check{scale}{2}{dir}{dir1};
      Y_parent_real = x_check{scale+1}{1}{dir}{dir1};
      Y_parent_imag = x_check{scale+1}{2}{dir}{dir1};
      Y_parent_real  = expand(Y_parent_real);
      Y_parent_imag  = expand(Y_parent_imag);
      
      Wsig = conv2(windowfilt, windowfilt, (Y_coef_real).^2, 'same');
      Ssig = sqrt(max(Wsig-Nsig.^2, eps));
      
      T = sqrt(3)*Nsig^2./Ssig;
      
      Y_coef = Y_coef_real + sqrt(-1)*Y_coef_imag;
      Y_parent = Y_parent_real + sqrt(-1)*Y_parent_imag;
      Y_coef = bishrink(Y_coef, Y_parent, T*lambda);
      
      x_check{scale}{1}{dir}{dir1} = real(Y_coef);
      x_check{scale}{2}{dir}{dir1} = imag(Y_coef);
    end
  end
end
