% 
% function Phi = MS_BCS_SPL_GenerateProjection(block_sizes, subrates, ...
%     filename, method)
% 
%   This function generates the block-based random projections for 
%   wavelet-domain block-based sampling. Both dense Gaussian matrices
%   as well as DCT-based structured random matrix sampling is
%   supported (method = 'gaussian' or 'dct', respectively).
%   The randomly generated sampling operator is stored in filename if
%   given so that the it need not be re-generated for subsequent
%   trials.
%

%
% MS-BCS-SPL: Multiscale Block Compressed Sensing - Smooth Projected Landweber
% Copyright (C) 2012  James E. Fowler
% http://www.ece.mstate.edu/~fowler
% 
% This program is free software; you can redistribute it and/or
% modify it under the terms of the GNU General Public License
% as published by the Free Software Foundation; either version 2
% of the License, or (at your option) any later version.
% 
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU General Public License for more details.
% 
% You should have received a copy of the GNU General Public License
% along with this program; if not, write to the Free Software
% Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
%

function Phi = MS_BCS_SPL_GenerateProjection(block_sizes, subrates, ...
    filename, method)

num_levels = length(block_sizes);
  
if (nargin < 4)
  method = 'gaussian';
elseif ((~strcmp(method, 'gaussian')) && (~strcmp(method,'dct')))
  disp('method must be either gaussian or dct');
  return;
end

if (strcmp(method, 'gaussian'))
  if ((nargin > 2) && exist(filename, 'file'))
    block_sizes2 = block_sizes;
    load(filename);
    if (any(block_sizes2 ~= block_sizes))
      clear('Phi');
      block_sizes = block_sizes2;
    end
  end
  
  if (~exist('Phi', 'var'))
    for level = 1:num_levels
      N = block_sizes(level) * block_sizes(level);
      Phi{level} = orth(randn(N, N))';
    end  
  end
  
  if ((nargin > 2) && (~exist(filename, 'file')))
    save(filename, 'Phi', 'block_sizes');
  end
  
  for level = 1:num_levels
    N = block_sizes(level) * block_sizes(level);
    M = round(subrates(level) * N);
    Phi{level} = Phi{level}(1:M, :);
  end
else
  if ((nargin > 2) && exist(filename, 'file'))
    block_sizes2 = block_sizes;
    load(filename);
    if (any(block_sizes2 ~= block_sizes))
      clear('selection');
      clear('permutation');
      block_sizes = block_sizes2;
    end
  end

  if (~exist('selection', 'var'))
    for level = 1:num_levels
      N = block_sizes(level) * block_sizes(level);
      permutation{level} = randperm(N)';
      selection{level} = randperm(N)';
    end    
  end    

  if ((nargin > 2) && (~exist(filename, 'file')))
    save(filename, 'permutation', 'selection', 'block_sizes');
  end
  
  for level = 1:num_levels
    N = block_sizes(level) * block_sizes(level);
    M = round(subrates(level) * N);
    selection{level} = selection{level}(1:M);
    A = @(x) ScrambledBlockDCTSample(x, selection{level}, permutation{level}, ...
	block_sizes(level));
    AT = @(y) ScrambledBlockDCTSampleAdjoint(y, selection{level}, ...
	permutation{level}, block_sizes(level));
    
    Phi{level} = { block_sizes(level) ; A ; AT };
  end  
end    


function y = ScrambledBlockDCTSample(x, selection, permutation, block_size)
num_blocks = size(x, 2);
x(permutation, :) = x;
x_blk = reshape(x, [block_size (num_blocks * block_size)]);
fx = dct(x_blk);
fx = reshape(fx, [(block_size * block_size) num_blocks]);
y = fx(selection, :);

function x = ScrambledBlockDCTSampleAdjoint(y, selection, ...
    permutation, block_size)
num_blocks =  size(y, 2);
fx = zeros(block_size * block_size, num_blocks);
fx(selection, :) = y;
fx = reshape(fx, [block_size (num_blocks * block_size)]);
x = idct(fx);
x = reshape(x, [(block_size * block_size) num_blocks]);
x = x(permutation, :);
