% 
% function reconstructed_image = ...
%     MS_GPSR_Decoder(y, Phi, subrates, num_rows, num_cols)
% 
%   This function performs wavelet-domain GPSR reconstruction of y.
%   Phi gives the projection matrix. The reconstructed
%   image, of size num_rows x num_cols, is returned as
%   reconstructed_image.
%

%
% MS-BCS-SPL: Multiscale Block Compressed Sensing - Smooth Projected Landweber
% Copyright (C) 2012  James E. Fowler
% http://www.ece.mstate.edu/~fowler
% 
% This program is free software; you can redistribute it and/or
% modify it under the terms of the GNU General Public License
% as published by the Free Software Foundation; either version 2
% of the License, or (at your option) any later version.
% 
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU General Public License for more details.
% 
% You should have received a copy of the GNU General Public License
% along with this program; if not, write to the Free Software
% Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
%

function reconstructed_image = ...
    MS_GPSR_Decoder(y, Phi, subrates, num_rows, num_cols)

num_levels = length(y) - 1;

if (isnumeric(Phi{1}))
  for level = 1:num_levels
    Phi_matrix = Phi{level};
    Phi_transpose{level} = @(y) Phi_matrix' * y;
    Phi{level} = @(x) Phi_matrix * x;
  end
elseif (iscell(Phi{1}))
  for level = 1:num_levels
    Phi_transpose{level} = Phi{level}{2};
    Phi{level} = Phi{level}{1};
  end
else
  disp('Incorrect Phi format');
  return;
end

tau = 0.02;

x_tilde = zeros(num_rows, num_cols);

subband_num_rows = num_rows / 2^num_levels;
subband_num_cols = num_cols / 2^num_levels;

x_tilde(1:subband_num_rows, 1:subband_num_cols) = ...
    reshape(y{1}, [subband_num_rows subband_num_cols]);
for level = 1:num_levels;
  if (subrates(level) == 1)
    x_gpsr = Phi_transpose{level}(y{level + 1});
  else
    x_gpsr = GPSR_BB(y{level + 1}, Phi{level}, tau, ...
	'AT', Phi_transpose{level}, 'Debias', 1, ...
	'ToleranceA', 1e-4, 'Verbose', 0);
  end
  subband_length = subband_num_rows * subband_num_cols;
  subband1 = x_gpsr(1:subband_length);
  subband2 = x_gpsr((subband_length + 1):(2 * subband_length));
  subband3 = x_gpsr((2 * subband_length + 1):(3 * subband_length));

  x_tilde(1:subband_num_rows, ...
      (subband_num_cols + 1):(2 * subband_num_cols)) = ...
      reshape(subband1, [subband_num_rows subband_num_cols]);
  x_tilde((subband_num_rows + 1):(2 * subband_num_rows), ...
      1:subband_num_cols) = ...
      reshape(subband2, [subband_num_rows subband_num_cols]);
  x_tilde((subband_num_rows + 1):(2 * subband_num_rows), ...
      (subband_num_cols + 1):(2 * subband_num_cols)) = ...
      reshape(subband3, [subband_num_rows subband_num_cols]);

  subband_num_rows = subband_num_rows * 2;
  subband_num_cols = subband_num_cols * 2;
end  

reconstructed_image = waveletcdf97(x_tilde, -num_levels);
