% 
% function Phi = MS_GPSR_GenerateProjection(subband_sizes, subrates, ...
%     filename, method)
% 
%   This function generates the random projections for 
%   wavelet-domain sampling. Both dense Gaussian matrices
%   as well as DCT-based structured random matrix sampling is
%   supported (method = 'gaussian' or 'dct', respectively).
%   The randomly generated sampling operator is stored in filename if
%   given so that the it need not be re-generated for subsequent
%   trials.
%

%
% MS-BCS-SPL: Multiscale Block Compressed Sensing - Smooth Projected Landweber
% Copyright (C) 2012  James E. Fowler
% http://www.ece.mstate.edu/~fowler
% 
% This program is free software; you can redistribute it and/or
% modify it under the terms of the GNU General Public License
% as published by the Free Software Foundation; either version 2
% of the License, or (at your option) any later version.
% 
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU General Public License for more details.
% 
% You should have received a copy of the GNU General Public License
% along with this program; if not, write to the Free Software
% Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
%

function Phi = MS_GPSR_GenerateProjection(subband_sizes, subrates, ...
    filename, method)

num_levels = length(subrates);
  
if (nargin < 4)
  method = 'gaussian';
elseif ((~strcmp(method, 'gaussian')) && (~strcmp(method,'dct')))
  disp('method must be either gaussian or dct');
  return;
end

if (strcmp(method, 'gaussian'))
  if ((nargin > 2) && exist(filename, 'file'))
    load(filename);
    for level = 1:num_levels
      if (size(Phi{level}, 1) ~= 3 * prod(subband_sizes{level}))
	disp('Creating new Phi...');
	clear('Phi');
	break;
      end
    end
  end
  
  if (~exist('Phi', 'var'))
    for level = 1:num_levels
      N = 3 * prod(subband_sizes{level});
      Phi{level} = orth(randn(N, N))';
    end  
  end
  
  if ((nargin > 2) && (~exist(filename, 'file')))
    save(filename, 'Phi');
  end
  
  for level = 1:num_levels
    N = 3 * prod(subband_sizes{level});
    M = round(subrates(level) * N);
    Phi{level} = Phi{level}(1:M, :);
  end
else
  if ((nargin > 2) && exist(filename, 'file'))
    load(filename);
    for level = 1:num_levels
      if (length(permutation) ~= 3 * prod(subband_sizes{level}))
	clear('selection');
	clear('permutation');
	break;
      end
    end
  end

  if (~exist('selection', 'var'))
    for level = 1:num_levels
      N = 3 * prod(subband_sizes{level});
      permutation{level} = randperm(N)';
      selection{level} = randperm(N)';
    end    
  end    

  if ((nargin > 2) && (~exist(filename, 'file')))
    save(filename, 'permutation', 'selection');
  end
  
  for level = 1:num_levels
    N = 3 * prod(subband_sizes{level});
    M = round(subrates(level) * N);
    selection{level} = selection{level}(1:M);
    A = @(x) ScrambledBlockDCTSample(x, selection{level}, permutation{level}, ...
	32);
    AT = @(y) ScrambledBlockDCTSampleAdjoint(y, selection{level}, ...
	permutation{level}, 32);
    
    Phi{level} = { A ; AT };
  end  
end    


function y = ScrambledBlockDCTSample(x, selection, permutation, block_size)
num_blocks = length(x) / block_size;
x(permutation) = x;
x_blk = reshape(x, [block_size num_blocks]);
fx = dct(x_blk);
fx = fx(:);
y = fx(selection);

function x = ScrambledBlockDCTSampleAdjoint(y, selection, ...
    permutation, block_size)
num_blocks =  length(permutation) / block_size;
fx = zeros(length(permutation), 1);
fx(selection) = y;
fx = reshape(fx, [block_size num_blocks]);
x = idct(fx);
x = x(:);
x = x(permutation);
